from .node import Node
from soffosai.core.services import DocumentsIngestService, DocumentsSearchService, DocumentsDeleteService
from soffosai.core.services import inspect_arguments


class DocumentsIngestNode(Node):
    '''
    Ingest a text to Soffos' database, returns document_id that references that document
    '''
    def __init__(self, name:str, document_name:str, text:str=None, tagged_elements:list=None, meta:dict=None):
        source = inspect_arguments(self.__init__, name, document_name, text, tagged_elements, meta)
        source['name'] = source['document_name']
        service = DocumentsIngestService
        super().__init__(name, service, source)


class DocumentsSearchNode(Node):
    '''
    Return details about a document that was ingested to Soffos. 
    Takes document_ids, query, or filter.  Returns Passages of the documents and concatenated text.
    '''
    def __init__(self, name:str, query:str=None, filters:dict=None, document_ids:list=None, top_n_keywords:int=5,
        top_n_natural_language:int=5, date_from:str=None, date_until:str=None):
        source = inspect_arguments(self.__init__, name, query, filters, document_ids, top_n_keywords,
        top_n_natural_language, date_from, date_until)
        service = DocumentsSearchService
        super().__init__(name, service, source)


class DocumentsDeleteNode(Node):
    '''
    Deletes a document from the Soffos db that is referred to by the given document_ids
    '''
    def __init__(self, name:str, document_ids:list):
        source = inspect_arguments(self.__init__, name, document_ids)
        service = DocumentsDeleteService
        super().__init__(name, service, source)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-04-01
Purpose: The base Service class
-----------------------------------------------------
'''
import inspect
import soffosai
import json, io
import abc, requests, os, mimetypes, uuid
from soffosai.common.constants import SOFFOS_SERVICE_URL, FORM_DATA_REQUIRED
from soffosai.common.service_io_map import SERVICE_IO_MAP
from soffosai.common.serviceio_fields import ServiceIO


visit_docs_message = "Kindly visit https://platform.soffos.ai/playground/docs#/ for guidance."
input_structure_message = "To learn what the input dictionary should look like, access it by <your_service_instance>.input_structure"


def inspect_arguments(func, *args, **kwargs):
    '''
    Given a function, args and kwargs: 
    create a dictionary with keys as the arg names and values as the arg values
    '''
    # Get the list of argument names
    sig = inspect.signature(func)
    arg_names = list(sig.parameters.keys())

    # Combine positional arguments and keyword arguments
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    combined_args = bound_args.arguments

    # Filter out positional arguments not present in arg_names
    arguments = {name: combined_args[name] for name in arg_names if name in combined_args}
    unziped_kwargs = {}
    for key,value in arguments.items():
        if key != "kwargs":
            if value != None:
                unziped_kwargs[key] = value
        else:
            for key2, value2 in arguments['kwargs'].items():
                if value2 != None:
                    unziped_kwargs[key2] = value2
    
    if unziped_kwargs.get('name'):
        unziped_kwargs.pop('name')

    return unziped_kwargs


def format_uuid(uuid):
    formatted_uuid = '-'.join([
        uuid[:8],
        uuid[8:12],
        uuid[12:16],
        uuid[16:20],
        uuid[20:]
    ])
    return formatted_uuid


def is_valid_uuid(uuid_string):
    if isinstance(uuid_string, type): # coming from pipeline validation, always True because it will come from Soffos Services
        return True

    if "-" not in uuid_string:
        formatted_uuid = format_uuid(uuid_string)
    else:
        formatted_uuid = uuid_string
    try:
        uuid_obj = uuid.UUID(formatted_uuid)
    except ValueError:
        return False
    return str(uuid_obj) == formatted_uuid


class SoffosAIService:
    '''
    Base service class for all Soffos Services
    '''
    def __init__(self, service:str, **kwargs) -> None:            
        if kwargs.get("apikey"):
            apikey = kwargs['apikey']
        else:
            apikey = soffosai.api_key
            
        self.headers = {
            "x-api-key": apikey,
        }
        self._apikey = apikey
        self._service = service
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(service)
        self._payload = {}
        self._payload_keys = self._payload.keys()
        self._args_dict = {}


    @property
    def input_structure(self):
        '''
        These are the valid fields of the src dictionary for this service. Take note that some of the fields should not exist at the same time.
        To view fields that cannot co-exist, access the 'choose_one' property.
        '''
        return self._serviceio.input_structure
    

    @property
    def choose_one(self):
        '''
        These keys cannot co-exist in this service's src.
        '''
        return self._serviceio.require_one_of_choice


    def validate_payload(self, payload):
        '''
        checks if the input type is allowed for the service
        '''
        if not isinstance(payload, dict):
            raise TypeError("payload should be a dictionary")

        # check for missing arguments
        user_from_src = payload.get('user')
        if not user_from_src:
            return False, f"{self._service}: user key is required in the payload"

        if len(self._serviceio.required_input_fields) > 0:
            missing_requirements = []
            for required in self._serviceio.required_input_fields:
                if required not in payload:
                    missing_requirements.append(required)
            if len(missing_requirements) > 0:
                return False, f"{self._service}: Please provide {missing_requirements} on your payload. {visit_docs_message}. {input_structure_message}"
        
        if len(self._serviceio.require_one_of_choice) > 0:
            group_error = []
            for group in self._serviceio.require_one_of_choice:
                found_choices = []
                for choice in group:
                    if choice in payload:
                        found_choices.append(choice)
                if len(found_choices) == 0:
                    group_error.append(f"{self._service}: Please provide one of these values on your payload: {group}")
                elif len(found_choices) > 1:
                    group_error.append(f"{self._service}: Please only include one of these values: {group}")
            
            if len(group_error) > 0:
                return False, group_error

        # check if payload has proper type:
        input_structure = self._serviceio.input_structure
        value_errors = []
        for key, value in payload.items():
            if key in input_structure.keys():

                if not isinstance(input_structure[key], type):
                    input_type = type(input_structure[key])
                else:
                    input_type = input_structure[key]

                if not isinstance(value, input_type) and value != input_type: # the second condition is for pipeline
                    wrong_type = value if isinstance(value, type) else type(value)
                    value_errors.append(f"{key} requires {input_structure[key]} but {wrong_type} is provided.")
        
        special_validation_passed, error_on_special_validation = self._serviceio.special_validation(payload)
        if not special_validation_passed:
            value_errors.append(error_on_special_validation)
        if len(value_errors) > 0:
            return False, value_errors

        if "document_ids" in self._payload:
            if isinstance(payload['document_ids'], list):
                for _id in payload["document_ids"]:
                    valid_uuid = is_valid_uuid(_id)
                    if not valid_uuid:
                        return False, f"{_id} is invalid document_id"
        
        return True, None


    def get_data(self, payload):
        '''
        Prepare the json or form data input of the service
        '''
        
        request_data = {}
        for key, value in payload.items():
            if key != 'file':
                request_data[key] = value

        return request_data


    def handle_file(self, file_stream, filename, mime_type):
        # Using read(), seek() and other file like object operations on file_stream 
        buffer = file_stream.read()
        file_stream.seek(0)  # it's a good practice to reset the file pointer
        stream = io.BytesIO(buffer)
        file_tuple = (filename, stream, mime_type)
        return file_tuple


    def get_response(self, payload:dict={}, **kwargs) -> dict:
        '''
        Based on the knowledge/context, Soffos AI will now give you the data you need
        '''
        allow_input, message = self.validate_payload(payload)
        if "question" in payload.keys(): # the api receives the question as message.
            payload['message'] = payload['question']

        if not allow_input:
            raise ValueError(message)
        
        if not self._service:
            raise ValueError("Please provide the service you need from Soffos AI.")

        response = None
        data = self.get_data(payload)

        if self._service not in FORM_DATA_REQUIRED:
            self.headers["content-type"] = "application/json"
            try:
                response = requests.post(
                    url = SOFFOS_SERVICE_URL + self._service + "/",
                    headers = self.headers,
                    json = data,
                    timeout = 120
                )
                response.raise_for_status()
            except (requests.exceptions.HTTPError, requests.exceptions.ConnectionError, 
                    requests.exceptions.Timeout, requests.exceptions.RequestException) as err:
                return {
                    "status": 'Error',
                    "error": str(err)
                }
            
        else:
            file_obj = payload.get('file')
            if isinstance(file_obj, str):
                filename = str(os.path.basename(file_obj))
                mime_type, _ = mimetypes.guess_type(file_obj)
                with open(file_obj, 'rb') as file:
                    files = {
                        "file": (filename, file, mime_type)
                    }
                    try:
                        response = requests.post(
                            url = SOFFOS_SERVICE_URL + self._service + "/",
                            headers = self.headers,
                            data = data,
                            files = files,
                            timeout=120
                        )
                        response.raise_for_status()
                    except (requests.exceptions.HTTPError, requests.exceptions.ConnectionError, 
                            requests.exceptions.Timeout, requests.exceptions.RequestException) as err:
                        return {
                            "status": 'Error',
                            "error": str(err)
                        }

            else:
                filename = file_obj.name
                mime_type, _ = mimetypes.guess_type(filename)
                files = self.handle_file(file_obj, filename, mime_type)

                try:
                    response = requests.post(
                        url = SOFFOS_SERVICE_URL + self._service + "/",
                        headers = self.headers,
                        data = data,
                        files = files,
                        timeout=120
                    )
                    response.raise_for_status()
                except (requests.exceptions.HTTPError, requests.exceptions.ConnectionError, 
                        requests.exceptions.Timeout, requests.exceptions.RequestException) as err:
                    return {
                        "status": 'Error',
                        "error": str(err)
                    }
        
        if response.ok:
            return response.json()
        else:
            return {
                "status": response.status_code,
                "error": response.text
            }


    def __call__(self, payload:dict, **kwargs)->dict:
        return self.get_response(payload=payload,**kwargs)


    def __str__(self) -> str:
        return self._service


    @abc.abstractmethod
    def provide_output_type(self):
        '''
        Sends back the output datatype of the service
        '''
    

    @abc.abstractmethod
    def provide_source_type(self):
        '''
        Sends back the accepted source datatype of the service
        '''


    @abc.abstractmethod
    def get_default_output_key(self):
        '''
        Sends back the output type of the service
        '''



from .service_io import ServiceIO
from ..constants import ServiceString


class MicrolessonIO(ServiceIO):
    service = ServiceString.MICROLESSON
    required_input_fields = ["content"]
    input_structure = {
    "content": [
            {
            "source": str,
            "text": str
            },
        ],
    "user": "b5601df8-6af3-4c1a-9ded-b7df4c506eab"
    }
    output_structure = {
        "microlesson": str
    }



from .service_io import ServiceIO
from ..constants import ServiceString


class TagGenerationIO(ServiceIO):
    service = ServiceString.TAG_GENERATION
    required_input_fields = ["text"]
    optional_input_fields = ["types", "n"]
    input_structure = {
        "text": str,
        "types": [str, str, str], # can only take a subset of ["topic", "domain", "audience", "entity"]
        "n": int 
    }
    output_structure = {
        "tags": {
            "label1": [
                {
                    "tag": str,
                    "score": float
                },
                {
                    "tag": str,
                    "score": float
                }
            ],
            "label2": [
                {
                    "tag": str,
                    "score": float
                },
                {
                    "tag": str,
                    "score": float
                }
            ],
        }
    }



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-04-17
Purpose: Soffos Service base Node
-----------------------------------------------------
'''
from typing import Union
from soffosai.common.serviceio_fields import ServiceIO
from soffosai.core.services import SoffosAIService
from soffosai.common.constants import ServiceString


class Node:
    '''
    A SoffosAIService wrapper that holds information how the service is to be executed inside a 
    SoffosPipeline
    '''
    _service_io: ServiceIO

    def __init__(self, name, service:Union[ServiceString, SoffosAIService], source:dict={}) -> None:
        
        self._raw_service = service
        self.name = name
        self.source = source
        if isinstance(service, str):
            self.service:SoffosAIService = SoffosAIService(service=service)
        elif issubclass(service, SoffosAIService):
            self.service:SoffosAIService = service()
        else:
            raise ValueError("Upon initialization of the Node: invalid argument value for <service>.")
        

    def validate_node(self):
        '''
        Will check if the Node will run from the given source. Will also create the proper 
        source for the SoffosAIService
        '''
        validated_data = {}
        for key,value in self.source.items():
            if not isinstance(value, tuple):
                validated_data[key] = value
            else:
                raise ValueError(f"This source notation is only valid in a Pipeline. To execute a single node, provide the actual value for each source key")
            
        return validated_data


    def run(self, payload=None):
        if payload is not None:
            self.source = payload
            
        args = self.validate_node()

        return self.service.get_response(payload=payload)

    
    def __call__(self, payload={}, *args, **kwargs):
        '''
        This feature is only for testing/debugging a Node.
        To easily create and call a service, please use the SoffosAIService class
        '''
        self.service.get_response(payload=payload, *args, **kwargs)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-27
Purpose: Easily use Table Generator Service
-----------------------------------------------------
'''
from typing import Union
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString

TABLE_FORMATS = ['markdown', 'CSV']

class TableGeneratorService(SoffosAIService):
    '''
    The table generator module enables applications to extract numerical and statistical 
    data from raw text in a tabular format. For use-cases where data has to be manually 
    reviewed and cross-referenced, this module can bring enormous value.
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.TABLE_GENERATOR
        super().__init__(service, **kwargs)
    

    def __call__(self, user:str, text:str, table_format:str='markdown'):
        if table_format not in TABLE_FORMATS:
            raise ValueError(f"The argument table_format accepted values are: {TABLE_FORMATS}")
        payload = inspect_arguments(self.__call__, user, text, table_format)
        return super().__call__(payload)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-26
Purpose: Easily use Simplify Service
-----------------------------------------------------
'''
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString


class SimplifyService(SoffosAIService):
    '''
    Paraphrase and Simplify are available as two different flavors of the same module. 
    While the Paraphrase module attempts to change the wording while keeping the same level of complexity, 
    the Simplify module outputs more commonly used words without altering the meaning of the original text. 
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.SIMPLIFY
        super().__init__(service, **kwargs)
    

    def __call__(self, user:str, text:str):
        payload = inspect_arguments(self.__call__, user, text)
        return super().__call__(payload)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-27
Purpose: Easily use Profanity Service
-----------------------------------------------------
'''
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString


class ProfanityService(SoffosAIService):
    '''
    This module detects profanities and the level of offensiveness in a body of text.
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.PROFANITY
        super().__init__(service, **kwargs)
    

    def __call__(self, user:str, text:str):
        payload = inspect_arguments(self.__call__, user, text)
        return super().__call__(payload)



from .node import Node
from soffosai.core.services import inspect_arguments, SentimentAnalysisService


class SentimentAnalysisNode(Node):
    '''
    Sentiment Analysis Service configuration for Pipeline Use
    '''
    def __init__(self, name:str, text:str, sentence_split:int=4, sentence_overlap:bool=False):
        source = inspect_arguments(self.__init__, name, text, sentence_split, sentence_overlap)
        service = SentimentAnalysisService
        super().__init__(name, service, source)



from .node import Node
from soffosai.core.services import AmbiguityDetectionService, inspect_arguments


class AmbiguityDetectionNode(Node):
    '''
    Ambiguity Detection service configuration for Pipeline Use
    '''
    def __init__(self, name:str, text:str, sentence_split:int=4, sentence_overlap=False):
        service = AmbiguityDetectionService
        source = inspect_arguments(self.__init__, name, text, sentence_split, sentence_overlap)
        super().__init__(name, service, source)



from .service_io import ServiceIO
from ..constants import ServiceString


class TranscriptCorrectionIO(ServiceIO):
    service = ServiceString.TRANSCRIPTION_CORRECTION
    required_input_fields = ["text"]
    input_structure = {
        "text": str
    }
    output_structure = {
        "correction": str
    }



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-27
Purpose: Easily use Review Tagger Service
-----------------------------------------------------
'''
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString


class ReviewTaggerService(SoffosAIService):
    '''
    This module extracts key information from negative product reviews. It attempts to find 
    the referred object, it's fault and an action/verb that is associated with it. If any 
    of the information is not present, it returns "n/a". This is useful for organizations who 
    want to analyze product reviews in order to identify and prioritize the most important issues.
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.REVIEW_TAGGER
        super().__init__(service, **kwargs)
    

    def __call__(self, user:str, text:str):
        payload = inspect_arguments(self.__call__, user, text)
        return super().__call__(payload)



from .node import Node
from soffosai.core.services import inspect_arguments, QuestionAnsweringService


class QuestionAnsweringNode(Node):
    '''
    Question Answering Service configuration for Pipeline Use
    '''
    def __init__(self, name:str, question:str, document_text:str=None, document_ids:list=None, 
        check_ambiguity:bool=True, check_query_type:bool=True, generic_response:bool=False, meta:dict=None):

        source = inspect_arguments(self.__init__, name, question, document_text, document_ids, 
        check_ambiguity, check_query_type, generic_response, meta)
        source['message'] = question

        service = QuestionAnsweringService
        super().__init__(name, service, source)



from .service_io import ServiceIO
from ..constants import ServiceString


class QuestionAnsweringIO(ServiceIO):
    service = ServiceString.QUESTION_ANSWERING
    required_input_fields = ["question"] # api receives "message" but "question" is clearer in the sdk
    require_one_of_choice = [["document_text", "document_ids"]]
    defaults = ["document_text"]
    optional_input_fields = ["check_ambiguity", "check_query_type", "generic_responses"]
    input_structure = {
        "question": str,
        "document_ids": [
            str,
            str
        ],
        "document_text": str, # should not be defined if document_ids field is present
        "check ambiguity": bool,
        "check_query_type": bool,
        "generic_response": bool,
        "meta": {
            "session_id": str
        }
    }
    output_structure = {
        "answer": str,
        "valid_query": bool,
        "no_answer": bool,
        "message_id": str,
        "context": str,
        "highlights": [
            {
                "span": [int, int],
                "sentence": str
            },
        ],
        "passages": [dict, dict]
    }



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-30
Purpose: Define the standard Pipeline for converting, summarizing an ingested document
-----------------------------------------------------
'''
from soffosai.core import inspect_arguments
from soffosai.core.nodes import DocumentsSearchNode, SummarizationNode
from soffosai.core.pipelines import Pipeline

class DocumentSummaryPipeline(Pipeline):
    '''
    A Soffos Pipeline that takes document_ids, then summarizes the content.
    The output is a list containing the output object of file converter and summarization.
    '''
    def __init__(self, **kwargs) -> None:
        document_search_node = DocumentsSearchNode(
            name = "doc_search",
            document_ids= {"source": "user_input", "field": "document_ids"}
        )
        
        summarization_node = SummarizationNode(
            name = "summarization",
            text = {"source": "doc_search", "field": "text"},
            sent_length = {"source": "user_input", "field": "sent_length"}
        )

        nodes = [document_search_node, summarization_node]
        use_defaults = False
        super().__init__(nodes=nodes, use_defaults=use_defaults, **kwargs)


    def __call__(self, user, document_ids, sent_length):
        user_input = inspect_arguments(self.__call__, user, document_ids, sent_length)
        return super().__call__(user_input)



from .service_io import ServiceIO
from .ambiguity_detection_io import AmbiguityDetectionIO
from .answer_scoring_io import AnswerScoringIO
from .contradiction_detection_io import ContradictionDetectionIO
from .documents_io import DocumentsIngestIO, DocumentSearchIO, DocumentDeleteIO
from .email_analysis_io import EmailAnalysisIO
from .emotion_detection_io import EmotionDetectionIO
from .file_converter_io import FileConverterIO
from .language_detection_io import LanguageDetectionIO
from .lets_discuss_io import LetsDiscussCreateIO, LetsDiscussDeleteIO, LetsDiscussIO, LetsDiscussRetrieveIO
from .logical_error_detection_io import LogicalErrorDetectionIO
from .microlesson import MicrolessonIO
from .named_entity_recognition_io import NamedEntityRecognitionIO
from .paraphrase_io import ParaphraseIO
from .simplify_io import SimplifyIO
from .profanity_io import ProfanityIO
from .qna_generation_io import QuestionAndAnswerGenerationIO
from .question_answering_io import QuestionAnsweringIO
from .review_tagger_io import ReviewTaggerIO
from .sentiment_analysis_io import SentimentAnalysisIO
from .summarization_io import SummarizaionIO
from .table_generator_io import TableGeneratorIO
from .tag_generation_io import TagGenerationIO
from .transcript_correction import TranscriptCorrectionIO



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-26
Purpose: Easily use Documents Ingest, Search, and Delete Service
-----------------------------------------------------
'''
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString
from soffosai.common.service_io_map import SERVICE_IO_MAP
from soffosai.common.serviceio_fields import ServiceIO

class DocumentsIngestService(SoffosAIService):
    '''
    The Documents module enables ingestion of content into Soffos.
    Takes in the text and gives the document_id to reference the text in Soffos database
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.DOCUMENTS_INGEST
        super().__init__(service, **kwargs)
    
    def __call__(self, user:str, document_name:str, text:str=None, tagged_elements:list=None, meta:dict=None):
        payload = inspect_arguments(self.__call__, user, document_name, text, tagged_elements, meta)
        payload['name'] = document_name
        return super().__call__(payload)


class DocumentsSearchService(SoffosAIService):
    '''
    The Documents module enables search of ingested contents from Soffos.
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.DOCUMENTS_SEARCH
        super().__init__(service, **kwargs)
    

    def __call__(self, user:str, query:str=None, filters:dict=None, document_ids:list=None, top_n_keywords:int=5,
        top_n_natural_language:int=5, date_from:str=None, date_until:str=None):
        payload = inspect_arguments(self.__call__, user, query, filters, document_ids, top_n_keywords,
        top_n_natural_language, date_from, date_until)
        response:dict = super().__call__(payload)
        text = ""
        for passage in response.get('passages'):
            text = text + passage
        response['text'] = text
        return response


class DocumentsDeleteService(SoffosAIService):
    '''
    The Documents module enables deletion of ingested contents from Soffos.
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.DOCUMENTS_DELETE
        super().__init__(service, **kwargs)
    
    def __call__(self, user:str, document_ids:list):
        payload = inspect_arguments(self.__call__, user, document_ids)
        return super().__call__(payload)


class DocumentsService(SoffosAIService):
    '''
    The Documents module enables ingestion of content into Soffos.
    User can ingest text and get the reference to it as document_id.
    Cal also Retrieve the context and delete the ingested text from Soffos db.
    '''
    def __init__(self,  **kwargs) -> None:
        service = ServiceString.DOCUMENTS_SEARCH
        super().__init__(service, **kwargs)
    

    def __call__(self, user:str, query:str=None, filters:dict=None, document_ids:list=None, top_n_keywords:int=5,
        top_n_natural_language:int=5, date_from:str=None, date_until:str=None):
        return self.search(user, query=query, filters=filters, document_ids=document_ids, top_n_keywords=top_n_keywords,
        top_n_natural_language=top_n_natural_language, date_from=date_from, date_until=date_until)

    
    def search(self, user:str, query:str=None, filters:dict=None, document_ids:list=None, top_n_keywords:int=5,
        top_n_natural_language:int=5, date_from:str=None, date_until:str=None):
        self._service = ServiceString.DOCUMENTS_SEARCH
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(self._service)
        payload = inspect_arguments(self.search, user, query, filters, document_ids, top_n_keywords,
        top_n_natural_language, date_from, date_until)
        return self.get_response(payload=payload)


    def ingest(self, user:str, document_name:str, text:str=None, tagged_elements:list=None, meta:dict=None):
        self._service = ServiceString.DOCUMENTS_INGEST
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(self._service)
        payload = inspect_arguments(self.ingest, user, document_name, text, tagged_elements, meta)
        payload['name'] = document_name
        return self.get_response(payload)

    
    def delete(self, user:str, document_ids:list):
        self._service = ServiceString.DOCUMENTS_DELETE
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(self._service)
        payload = inspect_arguments(self.delete, user, document_ids)
        return self.get_response(payload)


    def get_response(self, payload:dict, **kwargs) -> dict:
        response = super().get_response(payload, **kwargs)
        text = ""
        if response.get('passages'):
            for passage in response.get['passages']:
                text = text + passage
            response['text'] = text
        return response



from .node import Node
from .ambiguity_detection import AmbiguityDetectionNode
from .answer_scoring import AnswerScoringNode
from .contradiction_detection import ContradictionDetectionNode
from .documents import DocumentsIngestNode, DocumentsSearchNode, DocumentsDeleteNode
from .email_analysis import EmailAnalysisNode
from .emotion_detection import EmotionDetectionNode
from .file_converter import FileConverterNode
from .language_detection import LanguageDetectionNode
from .lets_discuss import LetsDiscussCreateNode, LetsDiscussNode, LetsDiscussRetrieveNode, LetsDiscussDeleteNode
from .logical_error_detection import LogicalErrorDetectionNode
from .microlesson import MicrolessonNode
from .named_entity_recognition import NamedEntityRecognitionNode
from .paraphrase import ParaphraseNode
from .qna_generation import QuestionAndAnswerGenerationNode
from .question_answering import QuestionAnsweringNode
from .review_tagger import ReviewTaggerNode
from .sentiment_analysis import SentimentAnalysisNode
from .simplify import SimplifyNode
from .summarization import SummarizationNode
from .table_generator import TableGeneratorNode
from .tag_generation import TagGenerationNode
from .transcript_correction import TranscriptCorrectionNode



from .pipeline import Pipeline
from .document_summary import DocumentSummaryPipeline
from .file_ingest import FileIngestPipeline
from .file_summary_ingest import FileSummaryIngestPipeline
from .file_summary import FileSummaryPipeline



from .service_io import ServiceIO
from ..constants import ServiceString


class NamedEntityRecognitionIO(ServiceIO):
    service = ServiceString.NER
    required_input_fields = ["text"]
    optional_input_fields = ["labels"]
    input_structure = {
        "text": str,
        "labels": dict
    }
    output_structure = {
        "named_entities": [
            {
                "span": [
                    int,
                    int
                ],
                "tag": str,
                "text": str
            },
            {
                "span": [
                    int,
                    int
                ],
                "tag": str,
                "text": str
            },
        ]
    }



from .node import Node
from soffosai.core.services import AnswerScoringService, inspect_arguments


class AnswerScoringNode(Node):
    '''
    Answer Scoring service configuration for Pipeline Use
    '''
    def __init__(self, name:str, context:str, question:str, user_answer:str, answer:str=None):
        service = AnswerScoringService
        source = inspect_arguments(self.__init__, name, context, question, user_answer, answer)
        super().__init__(name, service, source)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-27
Purpose: Easily use Tag Generation Service
-----------------------------------------------------
'''
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString


class TagGenerationService(SoffosAIService):
    '''
    This module can generate tags for a piece of text that can aid with content search in 
    certain use-cases. It allows to specify a number of tags to be generated for each of 
    the categories "topic", "domain", "audience", "entity".
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.TAG_GENERATION
        super().__init__(service, **kwargs)
    

    def __call__(self, user:str, text:str, types:list=["topic"], n:int=10):
        '''
        Note: List of types of keywords to extract. Supported types:

        topic: Tags relating to the subject matter of the text. 
        domain: Tags relating to the domain of the text. For example, "AI", or "Science fiction". 
            In some cases, domain tags might be similar to topic tags. 
        audience: Tags relating to the type of audience the text is intended for. 
        entity: Entities such as people, places, products, etc. mentioned in the text.
        '''
        for _type in types:
            if _type not in ["topic", "domain", "audience", "entity"]:
                raise ValueError(f'{self._service} types argument\'s elements can only be "topic", "domain", "audience" and/or "entity".')
        payload = inspect_arguments(self.__call__, user, text, types, n)
        return super().__call__(payload)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-29
Purpose: Define the standard Pipeline for converting then ingesting a file
-----------------------------------------------------
'''
from soffosai.core import inspect_arguments
from soffosai.core.nodes import FileConverterNode, DocumentsIngestNode
from soffosai.core.pipelines import Pipeline

class FileIngestPipeline(Pipeline):
    '''
    A Soffos Pipeline that takes a file, convert it to its text content then saves it to Soffos db.
    the output is a list containing the output object of file converter and document ingest
    '''
    def __init__(self, **kwargs) -> None:


        file_converter_node = FileConverterNode(
            name = "fileconverter",
            file = {"source": "user_input", "field": "file"}
        )
        document_ingest_node = DocumentsIngestNode(
            name = "ingest",
            document_name = {"source": "user_input", "field": "file"},
            text = {"source": "fileconverter", "field": "text"}
        )

        nodes = [file_converter_node, document_ingest_node]
        use_defaults = False
        super().__init__(nodes=nodes, use_defaults=use_defaults, **kwargs)


    def __call__(self, user, file):
        user_input = inspect_arguments(self.__call__, user, file)
        return super().__call__(user_input)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-26
Purpose: Easily use NamedEntityRecognition Service
-----------------------------------------------------
'''
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString


class NamedEntityRecognitionService(SoffosAIService):
    '''
    Identifies named entities in text. It supports custom labels.
    This module is extremely versatile as the labels can be defined by the user. 
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.NER
        self.labels = {}
        super().__init__(service, **kwargs)
    

    def __call__(self, user:str, text:str, labels:dict=None):
        
        payload = inspect_arguments(self.__call__, user, text, labels)

        if not labels and len(self.labels.keys()) > 0:
            payload['labels'] = self.labels

        return super().__call__()


    def add_label(self, label:str, definition:str):
        '''
        Adds a TAG label and its description so that Soffos AI can identify the entities matching the tag
        '''
        self.labels[label] = definition
        


'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-26
Purpose: Easily use Emotion Detection Service
-----------------------------------------------------
'''
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString

_EMOTION_LIST = ["joy", "trust", "fear", "surprise", "sadness", "disgust", "anger", "anticipation"]

class EmotionDetectionService(SoffosAIService):
    '''
    Detect selected emotions within the provided text. The original text is chunked to 
    passages of a specified sentence length. Smaller chunks yield better accuracy.
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.EMOTION_DETECTION
        super().__init__(service, **kwargs)
    
    def __call__(self, user:str, text:str, sentence_split:int=4, sentence_overlap:bool=False, emotion_choices:list=_EMOTION_LIST):
        for emotion in emotion_choices:
            if emotion not in _EMOTION_LIST:
                raise ValueError(f"{emotion} is not valid as an emotion_choices element. Please choose from {_EMOTION_LIST}.")
        payload = inspect_arguments(self.__call__, user, text, sentence_split, sentence_overlap, emotion_choices)
        return super().__call__(payload)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-04-19
Purpose: Hold the output response of Soffos API in an object
-----------------------------------------------------
'''
import json
from soffosai.common.service_io_map import SERVICE_IO_MAP

class SoffosAiResponse:
    '''
    Holds the Soffos AI response
    '''
    def __init__(self, service, raw,**kwargs) -> None:
        self._raw:dict = raw
        self._context = self._raw.get("context")
        self._cost = self._raw.get("cost")
        self._charged_character_count = self._raw.get("charged_character_count")
        self._response = self._raw.get(SERVICE_IO_MAP[service].primary_output_key)
        self._tagged_elements = self._raw.get("tagged_elements")
        self._document_ids = self._raw.get("document_ids")
        other_document_ids = self.look_for_document_ids(self._raw)
        if len(other_document_ids) > 0:
            if self._document_ids:
                self._document_ids.append(other_document_ids)
            else:
                self._document_ids = [other_document_ids]
            
            self._raw['document_ids'] = self._document_ids
        
    @property
    def context(self):
        '''
        The context where the reply is based on
        '''
        return self._context

    @property
    def raw_response(self) -> dict:
        '''
        The raw json response from the ai but converted into a dictionary
        '''
        return self._raw

    @property
    def cost(self) -> dict:
        '''
        a dictionary describing the cost of the api call
        '''
        return self._cost

    @property
    def charged_character_count(self):
        '''
        The number of characters charged. It is based on input or output whichever is higher
        '''
        return self._charged_character_count

    @property
    def response(self) -> str or dict:
        '''
        The common response
        '''
        return self._response

    @property
    def tagged_elements(self):
        '''
        The tagged elements that can be used to optimize document ingestion.
        This is an output of the file converter and is useful for the document ingestion service
        '''
        return self._tagged_elements

    @property
    def document_ids(self):
        return self._document_ids
    
    def look_for_document_ids(self, raw:dict):
        document_ids = []
        if "document_id" in raw:
            document_ids.append(raw['document_id'])
        
        if "document_ids" in raw:
            document_ids.extend(raw['document_ids'])

        for key in raw.keys():
            if isinstance(raw[key], dict):
                if "document_id" in raw[key]:
                    document_ids.append(raw[key]['document_id'])
                if "document_ids" in raw[key]:
                    document_ids.extend(raw[key]['document_ids'])
        
        return document_ids


    def __str__(self) -> str:
        return self._response if isinstance(self._response, str) else json.dumps(self._response, indent=4)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-08-14
Purpose: Define the basic pipeline object
-----------------------------------------------------
'''
import soffosai
from soffosai.core.nodes.node import Node


def is_node_input(value):
    if not isinstance(value, dict):
        return False
    return "source" in value and "field" in value


class Pipeline:
    '''
    A controller for consuming multiple Services called stages.
    It validates all inputs of all stages before sending the first Soffos API request to ensure
    that the Pipeline will not waste credits.
    
    ** use_defaults=True means that stages will take input from the previous stages' 
    output of the same field name prioritizing the latest stage's output. 
    If the previous stages does not have it, it will take from the
    pipeline's user_input.  Also, the stages will only be supplied with the required fields + default
    of the require_one_of_choice fields.
    '''
    def __init__(self, nodes:list, use_defaults:bool=False, name=None, **kwargs) -> None:
        self._apikey = kwargs['apikey'] if kwargs.get('apikey') else soffosai.api_key
        self._stages = nodes
        
        self._input:dict = {}
        self._infos = []
        self._use_defaults = use_defaults
        self._execution_codes = []
        self._termination_codes = []

        error_messages = []
        if not isinstance(nodes, list):
            error_messages.append('stages field should be a list of Service Nodes')

        node_names = [node.name for node in nodes]
        for node in nodes:
            if not isinstance(node, Node) and not isinstance(node, Pipeline):
                error_messages.append(f'{node} is not an instance of Node.')
            
            if node_names.count(node.name) > 1:
                error_messages.append(f"Node name '{node.name}' is not unique.")

        if len(error_messages) > 0:
            raise ValueError("\\n".join(error_messages))

        # when a pipeline is used as another pipeline's input, it needs a name
        self.name = name

    
    def run(self, user_input):
        if not isinstance(user_input, dict):
            raise ValueError("User input should be a dictionary.")

        if "user" not in user_input:
            raise ReferenceError("'user' is not defined in the user_input.")

        if "text" in user_input:
            user_input['document_text'] = user_input['text']

        if "question" in user_input:
            user_input['message'] = user_input['question']

        if self._use_defaults:
            stages = self.set_defaults(self._stages, user_input)
        else:
            stages = self._stages

        self.validate_pipeline(stages, user_input)

        # termination referencing
        execution_code = user_input.get("execution_code")
        if execution_code:
            execution_code = self._apikey + execution_code
            if execution_code in self._execution_codes:
                raise ValueError("This execution code is still being used in an existing pipeline run.")
            else:
                self._execution_codes.append(execution_code)

        # Initialization of values
        infos = {}
        infos['user_input'] = user_input
        total_cost = 0.00

        # Execute per stage
        for stage in stages:
            # premature termination
            if execution_code in self._termination_codes:
                self._termination_codes.remove(execution_code)
                self._execution_codes.remove(execution_code)
                infos['total_cost'] = total_cost
                infos['warning'] = "This Soffos Pipeline has been prematurely terminated"
                return infos

            if isinstance(stage, Pipeline):
                stage: Pipeline
                response = stage.run(user_input)
                print(f"Response ready for {stage.name}.")
                pipe_output = {}
                for key, value in response.items():
                    if key != 'total_cost':
                        for subkey, subvalue in value.items():
                            pipe_output[subkey] = subvalue
                    else:
                        total_cost += value
                infos[stage.name] = pipe_output
                continue

            stage: Node
            # execute
            print(f"running {stage.service._service}.")
            tmp_source: dict = stage.source
            payload = {}
            for key, notation in tmp_source.items():
                # prepare payload
                if is_node_input(notation): # value is pointing to another node
                    value = infos[notation['source']][notation['field']]
                    if "pre_process" in notation:
                        if callable(notation['pre_process']):
                            payload[key] = notation['pre_process'](value)
                        else:
                            raise ValueError(f"{stage.name}: pre_process value should be a function.")
                    
                    else:
                        payload[key] = value
                else:
                    payload[key] = notation

            if 'user' not in payload:
                payload["user"] = user_input['user']
            
            payload['apikey'] = self._apikey

            response = stage.service.get_response(payload)
            if "error" in response:
                raise ValueError(response)
            
            print(f"Response ready for {stage.name}")
            infos[stage.name] = response
            total_cost += response['cost']['total_cost']

        infos['total_cost'] = total_cost

        # remove this execution code from execution codes in effect:
        if execution_code:
            self._execution_codes.remove(execution_code)

        return infos


    def validate_pipeline(self, stages, user_input):
        '''
        Before running the first service, the Pipeline will validate all nodes if they will all be
        executed successfully with the exception of database and server issues.
        '''
        if not isinstance(user_input, dict):
            raise ValueError("User input should be a dictionary.")

        if "user" not in user_input:
            raise ReferenceError("'user' is not defined in the user_input.")

        if "text" in user_input:
            user_input['document_text'] = user_input['text']

        error_messages = []

        for stage in stages:
            if isinstance(stage, Pipeline):
                stage:Pipeline
                
                if stage._use_defaults:
                    sub_pipe_stages = stage.set_defaults(stage._stages, user_input)
                else:
                    sub_pipe_stages = stage._stages

                stage.validate_pipeline(sub_pipe_stages, user_input)
                continue

            # If stage is a Node:
            stage: Node
            serviceio = stage.service._serviceio
            # checking required_input_fields is already handled in the Node's constructor

            # check if require_one_of_choices is present and not more than one
            if len(serviceio.require_one_of_choice) > 0:
                group_errors = []
                for group in serviceio.require_one_of_choice:
                    found_choices = [choice for choice in group if choice in stage.source]
                    if not found_choices:
                        group_errors.append(f"{stage.name}: Please provide one of these values on your payload: {group}.")
                    elif len(found_choices) > 1:
                        group_errors.append(f"Please include only one of these values: {group}.")
                    
                if len(group_errors) > 0:
                    error_messages.append(". ".join(group_errors))
            
            # check if datatypes are correct:
            for key, notation in stage.source.items():
                try:
                    required_datatype = self.get_serviceio_datatype(stage.service._serviceio.input_structure[key])
                except KeyError: # If there are user_input fields not required by the pipeline
                    continue
                
                if is_node_input(notation):
                    if "pre_process" in notation:
                        continue # will not check for type if there is a helper function
                    
                    if notation['source'] == "user_input":
                        user_input_type = type(user_input[notation['field']])
                        if user_input_type != required_datatype:
                            error_messages.append(f"{stage.name}: {required_datatype} required on user_input '{key}' field but {user_input_type} is provided.")
                    else:
                        source_found = False
                        for subnode in stages:
                            subnode: Node
                            if notation['source'] == subnode.name:
                                source_found = True
                                if isinstance(subnode, Pipeline):
                                    break
                                output_datatype = self.get_serviceio_datatype(subnode.service._serviceio.output_structure[notation['field']])
                                if output_datatype != required_datatype:
                                    error_messages.append(f"On {stage.name} node: The input datatype required for field ${key} is {required_datatype}. This does not match the datatype to be given by node ${subnode.name}'s ${notation['field']} field which is ${output_datatype}.")
                                break
                        if not source_found:
                            error_messages.append(f"Cannot find node '{notation['source']}'")
                
                else:
                    if type(notation) != required_datatype:
                        error_messages.append(f"On {stage.name} node: {key} requires ${required_datatype} but ${type(notation)} is provided.")

        if len(error_messages) > 0:
            raise ValueError(error_messages)
        
        return True


    def add_node(self, node):
        if isinstance(node, Node) or isinstance(node, Pipeline):
            self._stages.append(node)
        else:
            raise ValueError(f"{node} is not a Node instance")

    
    def set_defaults(self, stages, user_input):
        defaulted_stages = []
        for i, stage in enumerate(stages):
            if isinstance(stage, Pipeline):
                continue

            stage: Node
            stage_source = {}
            required_keys = stage.service._serviceio.required_input_fields
            require_one_choices = stage.service._serviceio.require_one_of_choice
            if len(require_one_choices) > 0:
                for choices in require_one_choices:
                    required_keys.append(choices[0]) # the default argument is the first one
            
            for required_key in required_keys:
                check_key = stage.source.get(required_key)
                if check_key and check_key != "default":
                    stage_source[required_key] = check_key
                    continue

                found_input = False
                for j in range(i-1, -1, -1):
                    stage_for_output:Node = stages[j]
                    stage_for_output_output_fields = stage_for_output.service._serviceio.output_structure
                    if required_key in stage_for_output_output_fields:
                        stage_source[required_key] = {
                            "source": stage_for_output.name,
                            "field": required_key
                        }
                        found_input = True
                
                    # special considerations:
                    elif required_key == "context" and "text" in stage_for_output_output_fields:
                        stage_source["context"] = {
                            "source": stage_for_output.name,
                            "field": "text"
                        }
                        found_input = True

                    elif required_key == "document_text" and "text" in stage_for_output_output_fields:
                        stage_source["document_text"] = {
                            "source": stage_for_output.name,
                            "field": "text"
                        }
                        found_input = True
                    
                if not found_input:
                    if required_key in user_input:
                        stage_source[required_key] = user_input[required_key]
                        stage_source[required_key] = {
                            "source": "user_input",
                            "field": required_key
                        }
                    else:
                        raise ReferenceError(f"Please add {required_key} to user input. The previous Nodes' outputs do not provide this data.")

            defaulted_stage = Node(stage.name, stage.service, stage_source)
            defaulted_stages.append(defaulted_stage)
        
        return defaulted_stages

    
    def terminate(self, termination_code):
        if termination_code:
            self._termination_codes.append(self._apikey + termination_code)
            return {"message": f"Request to terminate job {termination_code} received."}

        return {"message": f"Request to terminate job is not valid (execution code missing)."}


    def get_serviceio_datatype(self, key):
        if isinstance(key, type):
            return key
        return type(key)
    
    def __call__(self, user_input):
        return self.run(user_input)


'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-26
Purpose: Easily use Answer Scoring Service
-----------------------------------------------------
'''
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString


class AnswerScoringService(SoffosAIService):
    '''
    This module will mark the user's answer based on the provided context, 
    the question and, optionally, the expected correct answer..
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.ANSWER_SCORING
        super().__init__(service, **kwargs)
    
    def __call__(self, user:str, context:str, question:str, user_answer:str, answer:str=None)->dict:
        payload = inspect_arguments(self.__call__, user, context, question, user_answer, answer)
        return super().__call__(payload)



from .node import Node
from soffosai.core.services import EmotionDetectionService, inspect_arguments

_EMOTION_LIST = ["joy", "trust", "fear", "surprise", "sadness", "disgust", "anger", "anticipation"]

class EmotionDetectionNode(Node):
    '''
    Emotion Detection configuration for Pipeline Use
    '''
    def __init__(self, name:str, text:str, sentence_split:int=4, sentence_overlap:bool=False, emotion_choices:list=_EMOTION_LIST):
        source = inspect_arguments(self.__init__, name, text, sentence_split, sentence_overlap, emotion_choices)
        service = EmotionDetectionService
        super().__init__(name, service, source)



from .service_io import ServiceIO
from ..constants import ServiceString


class EmotionDetectionIO(ServiceIO):
    service = ServiceString.EMOTION_DETECTION
    required_input_fields = ["text"]
    optional_input_fields = ["sentence_split", "sentence_overlap", "emotion_choices"]
    input_structure = {
        "sentence_split": int,
        "sentence_overlap": bool,
        "text": str,
        "emotion_choices": list
    }
    output_structure = {
        "spans": [
            {
                "detected_emotions": list,
                "text": str,
                "span_start": int,
                "span_end": int
            },
        ]
    }

    def special_validation(payload):
        if payload.get("sentence_split"):
            if payload["sentence_split"] < 1 and payload["sentence_split"] > 10:
                return [False, "sentence_split must be between 1 and 10."]
        return super().special_validation(payload)



from .node import Node
from soffosai.core.services import MicrolessonService, inspect_arguments


class MicrolessonNode(Node):
    '''
    Microlesson Service configuration for Pipeline Use
    '''
    def __init__(self, name:str, content:list):
        source = inspect_arguments(self.__init__, name, content)
        service = MicrolessonService
        super().__init__(name, service, source)



from .service_io import ServiceIO
from ..constants import ServiceString


class TableGeneratorIO(ServiceIO):
    service = ServiceString.TABLE_GENERATOR
    required_input_fields = ["table_format", "text"]
    input_structure = {
        "table_format": str, # markdown or CSV
        "text": str
    }
    output_structure = {
    "tables": [
            {
                "title": str,
                "table": str,
                "note": str
            },
        ]
    }



from .node import Node
from soffosai.core.services import NamedEntityRecognitionService, inspect_arguments


class NamedEntityRecognitionNode(Node):
    '''
    NER Service configuration for Pipeline Use
    '''
    def __init__(self, name:str, text:str, labels:dict=None):
        source = inspect_arguments(self.__init__, name, text, labels)
        service = NamedEntityRecognitionService
        super().__init__(name, service, source)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-04-17
Purpose: Soffos Services Objects
-----------------------------------------------------
'''

from .service import SoffosAIService, inspect_arguments
from .ambiguity_detection import AmbiguityDetectionService
from .answer_scoring import AnswerScoringService
from .contradiction_detection import ContradictionDetectionService
from .documents import DocumentsIngestService, DocumentsSearchService, DocumentsDeleteService, DocumentsService
from .email_analysis import EmailAnalysisService
from .emotion_detection import EmotionDetectionService
from .file_converter import FileConverterService
from .language_detection import LanguageDetectionService
from .lets_discuss import LetsDiscussService, LetsDiscussCreateService, LetsDiscussRetrieveService, LetsDiscussDeleteService
from .logical_error_detection import LogicalErrorDetectionService
from .microlesson import MicrolessonService
from .NER import NamedEntityRecognitionService
from .paraphrase import ParaphraseService
from .profanity import ProfanityService
from .qna_generation import QuestionAndAnswerGenerationService
from .question_answering import QuestionAnsweringService
from .review_tagger import ReviewTaggerService
from .sentiment_analysis import SentimentAnalysisService
from .simplify import SimplifyService
from .summarization import SummarizationService
from .table_generator import TableGeneratorService
from .tag_generation import TagGenerationService
from .transcript_correction import TranscriptCorrectionService



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-26
Purpose: Easily use Let's Discuss Service
-----------------------------------------------------
'''
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString
from soffosai.common.service_io_map import SERVICE_IO_MAP
from soffosai.common.serviceio_fields import ServiceIO


class LetsDiscussService(SoffosAIService):
    '''
    The Let's Discuss module allows the user to have a conversation with the AI about the content 
    provided by the user. The main difference between this module and the Question Answering module 
    is that Let's Discuss keeps a history of the interactions.
    '''
    def __init__(self,  **kwargs) -> None:
        service = ServiceString.LETS_DISCUSS
        super().__init__(service, **kwargs)


    def create(self, user:str, context:str):
        self._service = ServiceString.LETS_DISCUSS_CREATE
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(self._service)
        payload = inspect_arguments(self.create, user, context)
        return self.get_response(payload)
    

    def __call__(self, user:str, session_id:str, query:str):
        self._service = ServiceString.LETS_DISCUSS
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(self._service)
        payload = inspect_arguments(self.__call__, user, session_id, query)
        return super().__call__(payload)
    

    def retrieve_sessions(self, user:str, return_messages:bool):
        self._service = ServiceString.LETS_DISCUSS_RETRIEVE
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(self._service)
        payload = inspect_arguments(self.retrieve_sessions, user, return_messages)
        return self.get_response(payload)
    
    
    def delete(self, user:str, session_ids:list):
        self._service = ServiceString.LETS_DISCUSS_DELETE
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(self._service)
        payload = inspect_arguments(self.delete, user, session_ids)
        return self.get_response(payload)


class LetsDiscussCreateService(SoffosAIService):
    '''
    A separate class for LetsDiscuss service to be used for creating a session only.
    '''
    def __call__(self, user:str, context:str):
        self._service = ServiceString.LETS_DISCUSS_CREATE
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(self._service)
        payload = inspect_arguments(self.__call__, user, context)
        return super().__call__(payload)


class LetsDiscussRetrieveService(SoffosAIService):
    '''
    A separate class for LetsDiscuss service to be used for retrieving sessions only.
    '''
    def __call__(self, user:str, return_messages:bool):
        self._service = ServiceString.LETS_DISCUSS_RETRIEVE
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(self._service)
        payload = inspect_arguments(self.__call__, user, return_messages)
        return super().__call__(payload)


class LetsDiscussDeleteService(SoffosAIService):
    '''
    A separate class for LetsDiscuss service to be used for deleting sessions only.
    '''
    def __call__(self, user:str, session_ids:list):
        self._service = ServiceString.LETS_DISCUSS_DELETE
        self._serviceio:ServiceIO = SERVICE_IO_MAP.get(self._service)
        payload = inspect_arguments(self.__call__, user, session_ids)
        return super().__call__(payload)



'''
Copyright (c)2022 - Soffos.ai - All rights reserved
Created at: 2023-06-23
Purpose: Easily use Ambiguity Detection Service
-----------------------------------------------------
'''
from .service import SoffosAIService, inspect_arguments
from soffosai.common.constants import ServiceString


class AmbiguityDetectionService(SoffosAIService):
    '''
    A SoffosAIService that finds statements or sentences in text that are not coherent, 
    or can be interpreted in multiple ways while also taking in account the surrounding context.
    '''

    def __init__(self,  **kwargs) -> None:
        service = ServiceString.AMBIGUITY_DETECTION
        super().__init__(service, **kwargs)
    
    def __call__(self, user:str, text:str, sentence_split:int=4, sentence_overlap:bool=False) -> dict:
        payload = inspect_arguments(self.__call__, user, text, sentence_split, sentence_overlap)
        return super().__call__(payload)



from .service_io import ServiceIO
from ..constants import ServiceString


class EmailAnalysisIO(ServiceIO):
    service = ServiceString.EMAIL_ANALYSIS
    required_input_fields = ["text"]
    input_structure = {
        "text": str
    }
    output_structure = {
        "analysis": {
            "keypoints": [str, str],
            "topics": [str, str],
            "sender": str,
            "receiver": [str, str],
            "mentions": [str, str],
            "sentiment": str,
            "urgency": str,
            "dates": [str, str]
        }
    }




from .service_io import ServiceIO
from ..constants import ServiceString

class FileConverterIO(ServiceIO):
    service = ServiceString.FILE_CONVERTER
    required_input_fields = ["file"]
    optional_input_fields = ["normalize"]
    input_structure = {
        "file": str,
        "normalize": int
    }
    output_structure = {
        "text": str,
        "tagged_elements": [
            {
            "text": str,
            "tag": str
            },
        ]
    }
